{
  "cells": [
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "%matplotlib inline"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "\n# Example: Hilbert space approximation for Gaussian processes.\n\nThis example replicates a few of the models in the excellent case\nstudy by Aki Vehtari [1] (originally written using R and Stan).\nThe case study uses approximate Gaussian processes [2] to model the\nrelative number of births per day in the US from 1969 to 1988.\nThe Hilbert space approximation is way faster than the exact Gaussian\nprocesses because it circumvents the need for inverting the\ncovariance matrix.\n\nThe original case study presented by Aki also emphasizes the iterative\nprocess of building a Bayesian model, which is excellent as a pedagogical\nresource. Here, however, we replicate only 4 out of all the models available in [1].\nThere are a few minor differences in the mathematical details of our models,\nwhich we had to make in order for the chains to mix properly. We have clearly\ncommented on the places where our models are different.\n\n**References:**\n    1. Gelman, Vehtari, Simpson, et al (2020), `\"Bayesian workflow book - Birthdays\"\n       <https://avehtari.github.io/casestudies/Birthdays/birthdays.html>`.\n    2. Riutort-Mayol G, B\u00fcrkner PC, Andersen MR, et al (2020),\n       \"Practical hilbert space approximate bayesian gaussian processes for probabilistic programming\".\n\n<img src=\"file://../_static/img/examples/hsgp.png\" align=\"center\">\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "import argparse\nimport functools\nimport operator\nimport os\n\nimport matplotlib.pyplot as plt\nimport numpy as np\nimport pandas as pd\n\nimport jax\nimport jax.numpy as jnp\nfrom tensorflow_probability.substrates import jax as tfp\n\nimport numpyro\nfrom numpyro import deterministic, plate, sample\nimport numpyro.distributions as dist\nfrom numpyro.infer import MCMC, NUTS\n\n\n# --- utility functions\ndef load_data():\n    URL = \"https://raw.githubusercontent.com/avehtari/casestudies/master/Birthdays/data/births_usa_1969.csv\"\n    data = pd.read_csv(URL, sep=\",\")\n    day0 = pd.to_datetime(\"31-Dec-1968\")\n    dates = [day0 + pd.Timedelta(f\"{i}d\") for i in data[\"id\"]]\n    data[\"date\"] = dates\n    data[\"births_relative\"] = data[\"births\"] / data[\"births\"].mean()\n    return data\n\n\ndef save_samples(out_path, samples):\n    \"\"\"\n    Save dictionary of arrays using numpys compressed binary format\n    Fast reading and writing and efficient storage\n    \"\"\"\n    np.savez_compressed(out_path, **samples)\n\n\nclass UnivariateScaler:\n    \"\"\"\n    Standardizes the data to have mean 0 and unit standard deviation.\n    \"\"\"\n\n    def __init__(self):\n        self._mean = None\n        self._std = None\n\n    def fit(self, x):\n        self._mean = np.mean(x)\n        self._std = np.std(x)\n        return self\n\n    def transform(self, x):\n        return (x - self._mean) / self._std\n\n    def inverse_transform(self, x):\n        return x * self._std + self._mean\n\n\ndef _agg(*args, scaler=None):\n    \"\"\"\n    Custom function for aggregating the samples\n    and transforming back to the desired scale.\n    \"\"\"\n    total = functools.reduce(operator.add, args)\n    return (100 * scaler.inverse_transform(total)).mean(axis=0)\n\n\n# --- modelling functions\ndef modified_bessel_first_kind(v, z):\n    v = jnp.asarray(v, dtype=float)\n    return jnp.exp(jnp.abs(z)) * tfp.math.bessel_ive(v, z)\n\n\ndef spectral_density(w, alpha, length):\n    c = alpha * jnp.sqrt(2 * jnp.pi) * length\n    e = jnp.exp(-0.5 * (length ** 2) * (w ** 2))\n    return c * e\n\n\ndef diag_spectral_density(alpha, length, L, M):\n    \"\"\"spd for squared exponential kernel\"\"\"\n    sqrt_eigenvalues = jnp.arange(1, 1 + M) * jnp.pi / 2 / L\n    return spectral_density(sqrt_eigenvalues, alpha, length)\n\n\ndef phi(x, L, M):\n    \"\"\"\n    The first `M` eigenfunctions of the laplacian operator in `[-L, L]`\n    evaluated at `x`. These are used for the approximation of the\n    squared exponential kernel.\n    \"\"\"\n    m1 = (jnp.pi / (2 * L)) * jnp.tile(L + x[:, None], M)\n    m2 = jnp.diag(jnp.linspace(1, M, num=M))\n    num = jnp.sin(m1 @ m2)\n    den = jnp.sqrt(L)\n    return num / den\n\n\ndef diag_spectral_density_periodic(alpha, length, M):\n    \"\"\"\n    Not actually a spectral density but these are used in the same\n    way. These are simply the first `M` coefficients of the Taylor\n    expansion approximation for the periodic kernel.\n    \"\"\"\n    a = length ** (-2)\n    J = jnp.arange(1, M + 1)\n    q2 = (2 * alpha ** 2 / jnp.exp(a)) * modified_bessel_first_kind(J, a)\n    return q2\n\n\ndef phi_periodic(x, w0, M):\n    \"\"\"\n    Basis functions for the approximation of the periodic kernel.\n    \"\"\"\n    m1 = jnp.tile(w0 * x[:, None], M)\n    m2 = jnp.diag(jnp.linspace(1, M, num=M))\n    mw0x = m1 @ m2\n    return jnp.cos(mw0x), jnp.sin(mw0x)\n\n\n# --- models\nclass GP1:\n    \"\"\"\n    Long term trend Gaussian process\n    \"\"\"\n\n    def __init__(self):\n        self.x_scaler = UnivariateScaler()\n        self.y_scaler = UnivariateScaler()\n\n    def model(self, x, L, M, y=None):\n        # intercept\n        intercept = sample(\"intercept\", dist.Normal(0, 1))\n\n        # long term trend\n        \u03c1 = sample(\"\u03c1\", dist.LogNormal(-1.0, 1.0))\n        \u03b1 = sample(\"\u03b1\", dist.HalfNormal(1.0))\n        eigenfunctions = phi(x, L, M)\n        spd = jnp.sqrt(diag_spectral_density(\u03b1, \u03c1, L, M))\n\n        with plate(\"basis1\", M):\n            \u03b21 = sample(\"\u03b21\", dist.Normal(0, 1))\n\n        f1 = deterministic(\"f1\", eigenfunctions @ (spd * \u03b21))\n        \u03bc = deterministic(\"\u03bc\", intercept + f1)\n        \u03c3 = sample(\"\u03c3\", dist.HalfNormal(0.5))\n        with plate(\"n_obs\", x.shape[0]):\n            sample(\"y\", dist.Normal(\u03bc, \u03c3), obs=y)\n\n    def get_data(self):\n        data = load_data()\n        x = data[\"id\"].values\n        y = data[\"births_relative\"].values\n        self.x_scaler.fit(x)\n        self.y_scaler.fit(y)\n        xsd = jnp.array(self.x_scaler.transform(x))\n        ysd = jnp.array(self.y_scaler.transform(y))\n        return dict(\n            x=xsd,\n            y=ysd,\n            L=1.5 * max(xsd),\n            M=10,\n        )\n\n    def make_figure(self, samples):\n        data = load_data()\n        dates = data[\"date\"]\n        y = 100 * data[\"births_relative\"]\n        \u03bc = 100 * self.y_scaler.inverse_transform(samples[\"\u03bc\"]).mean(axis=0)\n\n        f = plt.figure(figsize=(15, 5))\n        plt.axhline(100, color=\"k\", lw=1, alpha=0.8)\n        plt.plot(dates, y, marker=\".\", lw=0, alpha=0.3)\n        plt.plot(dates, \u03bc, color=\"r\", lw=2)\n        plt.ylabel(\"Relative number of births\")\n        plt.xlabel(\"\")\n        return f\n\n\nclass GP2:\n    \"\"\"\n    Long term trend with year seasonality component.\n    \"\"\"\n\n    def __init__(self):\n        self.x_scaler = UnivariateScaler()\n        self.y_scaler = UnivariateScaler()\n\n    def model(self, x, w0, J, L, M, y=None):\n        intercept = sample(\"intercept\", dist.Normal(0, 1))\n\n        # long term trend\n        \u03c11 = sample(\"\u03c11\", dist.LogNormal(-1.0, 1.0))\n        \u03b11 = sample(\"\u03b11\", dist.HalfNormal(1.0))\n        eigenfunctions = phi(x, L, M)\n        spd = jnp.sqrt(diag_spectral_density(\u03b11, \u03c11, L, M))\n        with plate(\"basis\", M):\n            \u03b21 = sample(\"\u03b21\", dist.Normal(0, 1))\n\n        # year-periodic component\n        \u03c12 = sample(\"\u03c12\", dist.HalfNormal(0.1))\n        \u03b12 = sample(\"\u03b12\", dist.HalfNormal(1.0))\n        cosines, sines = phi_periodic(x, w0, J)\n        spd_periodic = jnp.sqrt(diag_spectral_density_periodic(\u03b12, \u03c12, J))\n        with plate(\"periodic_basis\", J):\n            \u03b22_cos = sample(\"\u03b22_cos\", dist.Normal(0, 1))\n            \u03b22_sin = sample(\"\u03b22_sin\", dist.Normal(0, 1))\n\n        f1 = deterministic(\"f1\", eigenfunctions @ (spd * \u03b21))\n        f2 = deterministic(\n            \"f2\", cosines @ (spd_periodic * \u03b22_cos) + sines @ (spd_periodic * \u03b22_sin)\n        )\n        \u03bc = deterministic(\"\u03bc\", intercept + f1 + f2)\n        \u03c3 = sample(\"\u03c3\", dist.HalfNormal(0.5))\n        with plate(\"n_obs\", x.shape[0]):\n            sample(\"y\", dist.Normal(\u03bc, \u03c3), obs=y)\n\n    def get_data(self):\n        data = load_data()\n        x = data[\"id\"].values\n        y = data[\"births_relative\"].values\n        self.x_scaler.fit(x)\n        self.y_scaler.fit(y)\n        xsd = jnp.array(self.x_scaler.transform(x))\n        ysd = jnp.array(self.y_scaler.transform(y))\n        w0 = 2 * jnp.pi / (365.25 / self.x_scaler._std)\n        return dict(\n            x=xsd,\n            y=ysd,\n            w0=w0,\n            J=20,\n            L=1.5 * max(xsd),\n            M=10,\n        )\n\n    def make_figure(self, samples):\n        data = load_data()\n        dates = data[\"date\"]\n        y = 100 * data[\"births_relative\"]\n        y_by_day_of_year = 100 * data.groupby(\"day_of_year2\")[\"births_relative\"].mean()\n        \u03bc = 100 * self.y_scaler.inverse_transform(samples[\"\u03bc\"]).mean(axis=0)\n        f1 = 100 * self.y_scaler.inverse_transform(samples[\"f1\"]).mean(axis=0)\n        f2 = 100 * self.y_scaler.inverse_transform(samples[\"f2\"]).mean(axis=0)\n\n        fig, axes = plt.subplots(1, 2, figsize=(15, 5))\n        axes[0].plot(dates, y, marker=\".\", lw=0, alpha=0.3)\n        axes[0].plot(dates, \u03bc, color=\"r\", lw=2, alpha=1, label=\"Total\")\n        axes[0].plot(dates, f1, color=\"C2\", lw=3, alpha=1, label=\"Trend\")\n\n        axes[0].set_ylabel(\"Relative number of births\")\n        axes[0].set_title(\"All time\")\n\n        axes[1].plot(\n            y_by_day_of_year.index, y_by_day_of_year, marker=\".\", lw=0, alpha=0.5\n        )\n        axes[1].plot(\n            y_by_day_of_year.index, f2[:366], color=\"r\", lw=2, label=\"Year seaonality\"\n        )\n        axes[1].set_ylabel(\"Relative number of births\")\n        axes[1].set_xlabel(\"Day of year\")\n        for ax in axes:\n            ax.axhline(100, color=\"k\", lw=1, alpha=0.8)\n            ax.legend()\n\n        return fig\n\n\nclass GP3:\n    \"\"\"\n    Long term trend with yearly seasonaly and slowly varying day-of-week effect.\n    \"\"\"\n\n    def __init__(self):\n        self.x_scaler = UnivariateScaler()\n        self.y_scaler = UnivariateScaler()\n\n    def model(self, x, day_of_week, w0, J, L, M, L3, M3, y=None):\n        intercept = sample(\"intercept\", dist.Normal(0, 1))\n\n        # long term trend\n        \u03c11 = sample(\"\u03c11\", dist.LogNormal(-1.0, 1.0))\n        \u03b11 = sample(\"\u03b11\", dist.HalfNormal(1.0))\n        eigenfunctions = phi(x, L, M)\n        spd = jnp.sqrt(diag_spectral_density(\u03b11, \u03c11, L, M))\n        with plate(\"basis\", M):\n            \u03b21 = sample(\"\u03b21\", dist.Normal(0, 1))\n\n        # year-periodic component\n        \u03c12 = sample(\"\u03c12\", dist.HalfNormal(0.1))\n        \u03b12 = sample(\"\u03b12\", dist.HalfNormal(1.0))\n        cosines, sines = phi_periodic(x, w0, J)\n        spd_periodic = jnp.sqrt(diag_spectral_density_periodic(\u03b12, \u03c12, J))\n        with plate(\"periodic_basis\", J):\n            \u03b22_cos = sample(\"\u03b22_cos\", dist.Normal(0, 1))\n            \u03b22_sin = sample(\"\u03b22_sin\", dist.Normal(0, 1))\n\n        # day of week effect\n        with plate(\"plate_day_of_week\", 6):\n            \u03b2_week = sample(\"\u03b2_week\", dist.Normal(0, 1))\n        # next enforce sum-to-zero -- this is slightly different from Aki's model,\n        # which instead imposes Monday's effect to be zero.\n        \u03b2_week = jnp.concatenate([jnp.array([-jnp.sum(\u03b2_week)]), \u03b2_week])\n\n        # long term variation of week effect\n        \u03b13 = sample(\"\u03b13\", dist.HalfNormal(0.1))\n        \u03c13 = sample(\"\u03c13\", dist.LogNormal(1.0, 1.0))  # prior: very long-term effect\n        eigenfunctions_3 = phi(x, L3, M3)\n        spd_3 = jnp.sqrt(diag_spectral_density(\u03b13, \u03c13, L3, M3))\n        with plate(\"week_trend\", M3):\n            \u03b23 = sample(\"\u03b23\", dist.Normal(0, 1))\n\n        # combine\n        f1 = deterministic(\"f1\", eigenfunctions @ (spd * \u03b21))\n        f2 = deterministic(\n            \"f2\", cosines @ (spd_periodic * \u03b22_cos) + sines @ (spd_periodic * \u03b22_sin)\n        )\n        g3 = deterministic(\"g3\", eigenfunctions_3 @ (spd_3 * \u03b23))\n        \u03bc = deterministic(\"\u03bc\", intercept + f1 + f2 + jnp.exp(g3) * \u03b2_week[day_of_week])\n        \u03c3 = sample(\"\u03c3\", dist.HalfNormal(0.5))\n        with plate(\"n_obs\", x.shape[0]):\n            sample(\"y\", dist.Normal(\u03bc, \u03c3), obs=y)\n\n    def get_data(self):\n        data = load_data()\n        x = data[\"id\"].values\n        y = data[\"births_relative\"].values\n        self.x_scaler.fit(x)\n        self.y_scaler.fit(y)\n        xsd = jnp.array(self.x_scaler.transform(x))\n        ysd = jnp.array(self.y_scaler.transform(y))\n        w0 = 2 * jnp.pi / (365.25 / self.x_scaler._std)\n        dow = jnp.array(data[\"day_of_week\"].values) - 1\n        return dict(\n            x=xsd,\n            day_of_week=dow,\n            w0=w0,\n            J=20,\n            L=1.5 * max(xsd),\n            M=10,\n            L3=1.5 * max(xsd),\n            M3=5,\n            y=ysd,\n        )\n\n    def make_figure(self, samples):\n        data = load_data()\n        dates = data[\"date\"]\n        y = 100 * data[\"births_relative\"]\n        y_by_day_of_year = 100 * (\n            data.groupby(\"day_of_year2\")[\"births_relative\"].mean()\n        )\n        year_days = y_by_day_of_year.index.values\n\n        \u03bc = samples[\"\u03bc\"]\n        intercept = samples[\"intercept\"][:, None]\n        f1 = samples[\"f1\"]\n        f2 = samples[\"f2\"]\n        g3 = samples[\"g3\"]\n        \u03b2_week = samples[\"\u03b2_week\"]\n        \u03b2_week = np.concatenate([-\u03b2_week.sum(axis=1)[:, None], \u03b2_week], axis=1)\n\n        fig, axes = plt.subplots(2, 2, figsize=(15, 8), sharey=False, sharex=False)\n        axes[0, 0].plot(dates, y, marker=\".\", lw=0, alpha=0.3)\n        axes[0, 0].plot(\n            dates,\n            _agg(\u03bc, scaler=self.y_scaler),\n            color=\"r\",\n            lw=0,\n            label=\"Total\",\n            marker=\".\",\n            alpha=0.5,\n        )\n        axes[0, 1].plot(dates, y, marker=\".\", lw=0, alpha=0.3)\n        axes[0, 1].plot(\n            dates, _agg(f1, scaler=self.y_scaler), color=\"r\", lw=2, label=\"Trend\"\n        )\n        axes[1, 0].plot(year_days, y_by_day_of_year, marker=\".\", lw=0, alpha=0.3)\n        axes[1, 0].plot(\n            year_days,\n            _agg(f2[:, :366], scaler=self.y_scaler),\n            color=\"r\",\n            lw=2,\n            label=\"Year seasonality\",\n        )\n        axes[1, 1].plot(dates, y, marker=\".\", lw=0, alpha=0.3)\n        for day in range(7):\n            dow_trend = (jnp.exp(g3).T * \u03b2_week[:, day]).T\n            fit = _agg(intercept, f1, dow_trend, scaler=self.y_scaler)\n            axes[1, 1].plot(dates, fit, lw=2, color=\"r\")\n\n        axes[0, 0].set_title(\"Total\")\n        axes[0, 1].set_title(\"Long term trend\")\n        axes[1, 0].set_title(\"Year seasonality\")\n        axes[1, 1].set_title(\"Weekly effects with long term trend\")\n        for ax in axes.flatten():\n            ax.axhline(100, color=\"k\", lw=1, alpha=0.8)\n            ax.legend()\n\n        return fig\n\n\nclass GP4:\n    \"\"\"\n    Long term trend with yearly seasonaly, slowly varying day-of-week effect,\n    and special day effect including floating special days.\n    \"\"\"\n\n    def __init__(self):\n        self.x_scaler = UnivariateScaler()\n        self.y_scaler = UnivariateScaler()\n\n    def model(\n        self,\n        x,\n        day_of_week,\n        day_of_year,\n        memorial_days_indicator,\n        labour_days_indicator,\n        thanksgiving_days_indicator,\n        w0,\n        J,\n        L,\n        M,\n        L3,\n        M3,\n        y=None,\n    ):\n        intercept = sample(\"intercept\", dist.Normal(0, 1))\n\n        # long term trend\n        \u03c11 = sample(\"\u03c11\", dist.LogNormal(-1.0, 1.0))\n        \u03b11 = sample(\"\u03b11\", dist.HalfNormal(1.0))\n        eigenfunctions = phi(x, L, M)\n        spd = jnp.sqrt(diag_spectral_density(\u03b11, \u03c11, L, M))\n        with plate(\"basis\", M):\n            \u03b21 = sample(\"\u03b21\", dist.Normal(0, 1))\n\n        # year-periodic component\n        \u03c12 = sample(\"\u03c12\", dist.HalfNormal(0.1))\n        \u03b12 = sample(\"\u03b12\", dist.HalfNormal(1.0))\n        cosines, sines = phi_periodic(x, w0, J)\n        spd_periodic = jnp.sqrt(diag_spectral_density_periodic(\u03b12, \u03c12, J))\n        with plate(\"periodic_basis\", J):\n            \u03b22_cos = sample(\"\u03b22_cos\", dist.Normal(0, 1))\n            \u03b22_sin = sample(\"\u03b22_sin\", dist.Normal(0, 1))\n\n        # day of week effect\n        with plate(\"plate_day_of_week\", 6):\n            \u03b2_week = sample(\"\u03b2_week\", dist.Normal(0, 1))\n        # next enforce sum-to-zero -- this is slightly different from Aki's model,\n        # which instead imposes Monday's effect to be zero.\n        \u03b2_week = jnp.concatenate([jnp.array([-jnp.sum(\u03b2_week)]), \u03b2_week])\n\n        # long term separation of week effects\n        \u03c13 = sample(\"\u03c13\", dist.LogNormal(1.0, 1.0))\n        \u03b13 = sample(\"\u03b13\", dist.HalfNormal(0.1))\n        eigenfunctions_3 = phi(x, L3, M3)\n        spd_3 = jnp.sqrt(diag_spectral_density(\u03b13, \u03c13, L3, M3))\n        with plate(\"week_trend\", M3):\n            \u03b23 = sample(\"\u03b23\", dist.Normal(0, 1))\n\n        # Finnish horseshoe prior on day of year effect\n        # Aki uses slab_df=100 instead, but chains didn't mix\n        # in our case for some reason, so we lowered it to 50.\n        slab_scale = 2\n        slab_df = 50\n        scale_global = 0.1\n        \u03c4 = sample(\"\u03c4\", dist.HalfCauchy(scale=scale_global * 2))\n        c_aux = sample(\"c_aux\", dist.InverseGamma(0.5 * slab_df, 0.5 * slab_df))\n        c = slab_scale * jnp.sqrt(c_aux)\n        with plate(\"plate_day_of_year\", 366):\n            \u03bb = sample(\"\u03bb\", dist.HalfCauchy(scale=1))\n            \u03bb_tilde = jnp.sqrt(c) * \u03bb / jnp.sqrt(c + (\u03c4 * \u03bb) ** 2)\n            \u03b24 = sample(\"\u03b24\", dist.Normal(loc=0, scale=\u03c4 * \u03bb_tilde))\n\n        # floating special days\n        \u03b25_labour = sample(\"\u03b25_labour\", dist.Normal(0, 1))\n        \u03b25_memorial = sample(\"\u03b25_memorial\", dist.Normal(0, 1))\n        \u03b25_thanksgiving = sample(\"\u03b25_thanksgiving\", dist.Normal(0, 1))\n\n        # combine\n        f1 = deterministic(\"f1\", eigenfunctions @ (spd * \u03b21))\n        f2 = deterministic(\n            \"f2\", cosines @ (spd_periodic * \u03b22_cos) + sines @ (spd_periodic * \u03b22_sin)\n        )\n        g3 = deterministic(\"g3\", eigenfunctions_3 @ (spd_3 * \u03b23))\n        \u03bc = deterministic(\n            \"\u03bc\",\n            intercept\n            + f1\n            + f2\n            + jnp.exp(g3) * \u03b2_week[day_of_week]\n            + \u03b24[day_of_year]\n            + \u03b25_labour * labour_days_indicator\n            + \u03b25_memorial * memorial_days_indicator\n            + \u03b25_thanksgiving * thanksgiving_days_indicator,\n        )\n        \u03c3 = sample(\"\u03c3\", dist.HalfNormal(0.5))\n        with plate(\"n_obs\", x.shape[0]):\n            sample(\"y\", dist.Normal(\u03bc, \u03c3), obs=y)\n\n    def _get_floating_days(self, data):\n        x = data[\"id\"].values\n        memorial_days = data.loc[\n            data[\"date\"].dt.month.eq(5)\n            & data[\"date\"].dt.weekday.eq(0)\n            & data[\"date\"].dt.day.ge(25),\n            \"id\",\n        ].values\n\n        labour_days = data.loc[\n            data[\"date\"].dt.month.eq(9)\n            & data[\"date\"].dt.weekday.eq(0)\n            & data[\"date\"].dt.day.le(7),\n            \"id\",\n        ].values\n        labour_days = np.concatenate((labour_days, labour_days + 1))\n\n        thanksgiving_days = data.loc[\n            data[\"date\"].dt.month.eq(11)\n            & data[\"date\"].dt.weekday.eq(3)\n            & data[\"date\"].dt.day.ge(22)\n            & data[\"date\"].dt.day.le(28),\n            \"id\",\n        ].values\n        thanksgiving_days = np.concatenate((thanksgiving_days, thanksgiving_days + 1))\n\n        md_indicators = np.zeros_like(x)\n        md_indicators[memorial_days - 1] = 1\n        ld_indicators = np.zeros_like(x)\n        ld_indicators[labour_days - 1] = 1\n        td_indicators = np.zeros_like(x)\n        td_indicators[thanksgiving_days - 1] = 1\n        return {\n            \"memorial_days_indicator\": md_indicators,\n            \"labour_days_indicator\": ld_indicators,\n            \"thanksgiving_days_indicator\": td_indicators,\n        }\n\n    def get_data(self):\n        data = load_data()\n        x = data[\"id\"].values\n        y = data[\"births_relative\"].values\n        self.x_scaler.fit(x)\n        self.y_scaler.fit(y)\n        xsd = jnp.array(self.x_scaler.transform(x))\n        ysd = jnp.array(self.y_scaler.transform(y))\n        w0 = 2 * jnp.pi / (365.25 / self.x_scaler._std)\n        dow = jnp.array(data[\"day_of_week\"].values) - 1\n        doy = jnp.array((data[\"day_of_year2\"] - 1).values)\n        return dict(\n            x=xsd,\n            day_of_week=dow,\n            day_of_year=doy,\n            w0=w0,\n            J=20,\n            L=1.5 * max(xsd),\n            M=10,\n            L3=1.5 * max(xsd),\n            M3=5,\n            y=ysd,\n            **self._get_floating_days(data),\n        )\n\n    def make_figure(self, samples):\n        special_days = {\n            \"Valentine's\": pd.to_datetime(\"1988-02-14\"),\n            \"Leap day\": pd.to_datetime(\"1988-02-29\"),\n            \"Halloween\": pd.to_datetime(\"1988-10-31\"),\n            \"Christmas eve\": pd.to_datetime(\"1988-12-24\"),\n            \"Christmas day\": pd.to_datetime(\"1988-12-25\"),\n            \"New year\": pd.to_datetime(\"1988-01-01\"),\n            \"New year's eve\": pd.to_datetime(\"1988-12-31\"),\n            \"April 1st\": pd.to_datetime(\"1988-04-01\"),\n            \"Independence day\": pd.to_datetime(\"1988-07-04\"),\n            \"Labour day\": pd.to_datetime(\"1988-09-05\"),\n            \"Memorial day\": pd.to_datetime(\"1988-05-30\"),\n            \"Thanksgiving\": pd.to_datetime(\"1988-11-24\"),\n        }\n        \u03b24 = samples[\"\u03b24\"]\n        \u03b25_labour = samples[\"\u03b25_labour\"]\n        \u03b25_memorial = samples[\"\u03b25_memorial\"]\n        \u03b25_thanksgiving = samples[\"\u03b25_thanksgiving\"]\n\n        day_effect = np.array(\u03b24)\n        md_idx = special_days[\"Memorial day\"].day_of_year - 1\n        day_effect[:, md_idx] = day_effect[:, md_idx] + \u03b25_memorial\n        ld_idx = special_days[\"Labour day\"].day_of_year - 1\n        day_effect[:, ld_idx] = day_effect[:, ld_idx] + \u03b25_labour\n        td_idx = special_days[\"Thanksgiving\"].day_of_year - 1\n        day_effect[:, td_idx] = day_effect[:, td_idx] + \u03b25_thanksgiving\n        day_effect = 100 * day_effect.mean(axis=0)\n\n        fig = plt.figure(figsize=(12, 5))\n        plt.plot(np.arange(1, 367), day_effect)\n        for name, day in special_days.items():\n            xs = day.day_of_year\n            ys = day_effect[day.day_of_year - 1]\n            plt.plot(xs, ys, marker=\"o\", mec=\"k\", c=\"none\", ms=10)\n            plt.text(xs - 3, ys, name, horizontalalignment=\"right\")\n        plt.title(\"Special day effect\")\n        plt.ylabel(\"Relative number of births\")\n        plt.xlabel(\"Day of year\")\n        plt.xlim([-40, None])\n        return fig\n\n\ndef parse_arguments():\n    parser = argparse.ArgumentParser(description=\"Hilbert space approx for GPs\")\n    parser.add_argument(\"--num-samples\", nargs=\"?\", default=1000, type=int)\n    parser.add_argument(\"--num-warmup\", nargs=\"?\", default=1000, type=int)\n    parser.add_argument(\"--num-chains\", nargs=\"?\", default=1, type=int)\n    parser.add_argument(\n        \"--model\",\n        nargs=\"?\",\n        default=\"tywd\",\n        help=\"one of\"\n        '\"t\" (Long term trend),'\n        '\"ty\" (t + year seasonality),'\n        '\"tyw\" (t + y + slowly varying weekday effect),'\n        '\"tywd\" (t + y + w + special days effect)',\n    )\n    parser.add_argument(\"--device\", default=\"cpu\", type=str, help='use \"cpu\" or \"gpu\".')\n    parser.add_argument(\"--x64\", action=\"store_true\", help=\"Enable float64 precision\")\n    parser.add_argument(\n        \"--save-samples\",\n        default=\"\",\n        type=str,\n        help=\"Path where to store the samples. Must be '.npz' file.\",\n    )\n    parser.add_argument(\n        \"--save-figure\",\n        default=\"\",\n        type=str,\n        help=\"Path where to save the plot with matplotlib.\",\n    )\n    args = parser.parse_args()\n    return args\n\n\nNAME_TO_MODEL = {\n    \"t\": GP1,\n    \"ty\": GP2,\n    \"tyw\": GP3,\n    \"tywd\": GP4,\n}\n\n\ndef main(args):\n    is_sphinxbuild = \"NUMPYRO_SPHINXBUILD\" in os.environ\n    model = NAME_TO_MODEL[args.model]()\n    data = model.get_data()\n    mcmc = MCMC(\n        NUTS(model.model),\n        num_warmup=args.num_warmup,\n        num_samples=args.num_samples,\n        num_chains=args.num_chains,\n        progress_bar=False if is_sphinxbuild else True,\n    )\n    mcmc.run(jax.random.PRNGKey(0), **data)\n    if not is_sphinxbuild:\n        mcmc.print_summary()\n    posterior_samples = mcmc.get_samples()\n    if args.save_samples:\n        print(f\"Saving samples at {args.save_samples}\")\n        save_samples(args.save_samples, posterior_samples)\n    if args.save_figure:\n        print(f\"Saving figure at {args.save_figure}\")\n        fig = model.make_figure(posterior_samples)\n        fig.savefig(args.save_figure)\n        plt.close()\n\n    return model, data, mcmc, posterior_samples\n\n\nif __name__ == \"__main__\":\n    args = parse_arguments()\n    if args.x64:\n        numpyro.enable_x64()\n\n    numpyro.set_platform(args.device)\n    numpyro.set_host_device_count(args.num_chains)\n    main(args)"
      ]
    }
  ],
  "metadata": {
    "kernelspec": {
      "display_name": "Python 3",
      "language": "python",
      "name": "python3"
    },
    "language_info": {
      "codemirror_mode": {
        "name": "ipython",
        "version": 3
      },
      "file_extension": ".py",
      "mimetype": "text/x-python",
      "name": "python",
      "nbconvert_exporter": "python",
      "pygments_lexer": "ipython3",
      "version": "3.7.9"
    }
  },
  "nbformat": 4,
  "nbformat_minor": 0
}